#Cohen's kappa and Cramer's v for MTD(p) models




#Example with negative dependencies

#Transition matrix Q for certain 3-state Markov chain:

Q <- rbind(c(0.1,0.05,0.6), c(0.85,0.05,0.25), c(0.05,0.9,0.15))
nostates <- dim(Q)[1] #3

#Corresponding stationary marginal distribution:
pmarg <- eigen(Q, symmetric = FALSE)$vectors[,1]
pmarg <- Re(pmarg/sum(pmarg))
pmarg
#0.2734742 0.3450704 0.3814554

#Gini index:
nostates/(nostates-1) * (1-sum(pmarg^2))
#0.9909451

#Entropy:
-1/log(nostates) * sum(pmarg*log(pmarg))
#0.9915797



#To illustrate the transition matrix, one may draw a weighted directed graph:
#setInternet2(TRUE)
#install.packages("qgraph")
library("qgraph")

qgraph(t(Q), directed=TRUE, weighted=TRUE, labels=c(expression(s[0]),expression(s[1]),expression(s[2])), edge.color="black", edge.width=t(Q), esize=10, asize=10, fade=FALSE, layout="circle")





#MTD(1) model, i.e., Markov chain with Q, pmarg:


#Compute kappa and v for the Markov chain:
maxlag <- 15
cohenmc <- rep(0,maxlag)
vmc <- rep(0,maxlag)

sum2mc <- sum(pmarg^2)

indprobmc <- pmarg %*% t(pmarg)

Pk <- diag(1,nostates)

for(k in c(1:maxlag)){
	Pk <- Pk %*% Q #powers of transition matrix
	bivprob <- Pk %*% diag(pmarg)
	cohenmc[k] <- sum(diag(bivprob)-pmarg^2)/(1-sum2mc)
	vmc[k] <- sqrt(sum((bivprob-indprobmc)^2/indprobmc)/(nostates-1))
}
cohenmc
# [1] -0.3595820283 -0.1782925802  0.2794250783 -0.1189317453 -0.0369005619  0.0769710438 -0.0380124891 -0.0064888065  0.0208875349 -0.0118310876 -0.0006998787  0.0055772827 -0.0036031501  0.0001239735  0.0014625731

#Plot of Cohen's kappa:
plot(cohenmc, type="h", xlab = "k", ylab = expression(paste("Cohen's   ",kappa(k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)


#Compute "partial Cohen's kappa" recursively:
cohenpart <- rep(0,maxlag)
cohenpart[1] <- cohenmc[1]

a <- array(0, c(2,maxlag))
a[1,1] <- cohenmc[1]

for(k in c(1:(maxlag-1))){
	a[2,k+1] <- (cohenmc[k+1] - a[1,1:k] %*% cohenmc[k:1]) / (1 - a[1,1:k] %*% cohenmc[1:k])
	for(j in c(1:k)){
		a[2,j] <- a[1,j]-a[2,k+1]*a[1,k-j+1]
	}
	cohenpart[k+1] <- a[2,k+1]
	a[1,] <- a[2,]
}
plot(cohenpart, type="h", xlab = "k", ylab = expression(paste("Cohen's   ",kappa["part"](k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)


vmc
# [1] 0.676947144 0.445809537 0.282150185 0.189956698 0.126146016 0.079707368 0.053287332 0.035660041 0.022542564 0.014948078 0.010069764 0.006381377 0.004194358 0.002840238 0.001807708

#Plot of Cramer's v:
plot(vmc, type="h", xlab = "k", ylab = "Cramer's   v(k)", lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)

#Compute "partial Cramer's v" recursively:
cramerpart <- rep(0,maxlag)
cramerpart[1] <- vmc[1]

a <- array(0, c(2,maxlag))
a[1,1] <- vmc[1]

for(k in c(1:(maxlag-1))){
	a[2,k+1] <- (vmc[k+1] - a[1,1:k] %*% vmc[k:1]) / (1 - a[1,1:k] %*% vmc[1:k])
	for(j in c(1:k)){
		a[2,j] <- a[1,j]-a[2,k+1]*a[1,k-j+1]
	}
	cramerpart[k+1] <- a[2,k+1]
	a[1,] <- a[2,]
}
plot(cramerpart, type="h", xlab = "k", ylab = expression(paste("Cramer's   ",v["part"](k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)




#MTD(2) model
lambda1 <- 0.2
lambda2 <- 1-lambda1


#Compute kappa and v for the Markov chain:
maxlag <- 15
cohenmtd2 <- rep(0,maxlag)
vmtd2 <- rep(0,maxlag)

sum2mc <- sum(pmarg^2)

indprobmc <- pmarg %*% t(pmarg)


#Solve YW-equations for MTD(2) using vec and Kronecker product:
P0 <- diag(pmarg) #Initialized as P(0)

#Right-hand side of equation:
rhs <- lambda1*(Q %*% P0) + lambda1*lambda2*(Q %*% t(P0) %*% t(Q))

#Dimension of vec computations:
dim.vec <- prod(dim(Q))

#Coefficient matrix for left-hand side:
lhs.coef <- diag(1, dim.vec) - lambda2^2*kronecker(Q,Q)
P1 <-  matrix( solve(lhs.coef) %*% as.vector(rhs), ncol=3) #vec of P(1)

#Cross-checking YW equation:
P1 - (lambda1 * Q %*% P0) - (lambda2 * Q %*% t(P1))


Pk1 <- P0 #Initialized as P(0)
Pk <- P1 #Initialized as P(1)
bivprob <- Pk #Pk already contains bivariate probabilities
cohenmtd2[1] <- sum(diag(bivprob)-pmarg^2)/(1-sum2mc)
vmtd2[1] <- sqrt(sum((bivprob-indprobmc)^2/indprobmc)/(nostates-1))

for(k in c(2:maxlag)){
	Pk2 <- Pk1
	Pk1 <- Pk
	Pk <- (lambda1 * Q %*% Pk1) + (lambda2 * Q %*% Pk2)
	bivprob <- Pk #Pk already contains bivariate probabilities
	cohenmtd2[k] <- sum(diag(bivprob)-pmarg^2)/(1-sum2mc)
	vmtd2[k] <- sqrt(sum((bivprob-indprobmc)^2/indprobmc)/(nostates-1))
}
cohenmtd2
#0.0016903358 -0.3052185648 -0.0963102432 -0.0857770006  0.1091319465  0.1314804507 -0.0430907284 -0.0612685394 -0.0146752577  0.0071646013  0.0277566341  0.0108590809 -0.0144886187 -0.0100304660  0.0006342607

#Plot of Cohen's kappa:
plot(cohenmtd2, type="h", xlab = "k", ylab = expression(paste("Cohen's   ",kappa(k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)


#Compute "partial Cohen's kappa" recursively:
cohenpart <- rep(0,maxlag)
cohenpart[1] <- cohenmtd2[1]

a <- array(0, c(2,maxlag))
a[1,1] <- cohenmtd2[1]

for(k in c(1:(maxlag-1))){
	a[2,k+1] <- (cohenmtd2[k+1] - a[1,1:k] %*% cohenmtd2[k:1]) / (1 - a[1,1:k] %*% cohenmtd2[1:k])
	for(j in c(1:k)){
		a[2,j] <- a[1,j]-a[2,k+1]*a[1,k-j+1]
	}
	cohenpart[k+1] <- a[2,k+1]
	a[1,] <- a[2,]
}
plot(cohenpart, type="h", xlab = "k", ylab = expression(paste("Cohen's   ",kappa["part"](k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)



vmtd2
#0.15787101 0.54124604 0.14890837 0.28154697 0.10955140 0.13957667 0.07596777 0.07191869 0.04960085 0.03756933 0.02955817 0.01927832 0.01763178 0.01017519 0.01053310

#Plot of Cramer's v:
plot(vmtd2, type="h", xlab = "k", ylab = "Cramer's   v(k)", lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)


#Compute "partial Cramer's v" recursively:
cramerpart <- rep(0,maxlag)
cramerpart[1] <- vmtd2[1]

a <- array(0, c(2,maxlag))
a[1,1] <- vmtd2[1]

for(k in c(1:(maxlag-1))){
	a[2,k+1] <- (vmtd2[k+1] - a[1,1:k] %*% vmtd2[k:1]) / (1 - a[1,1:k] %*% vmtd2[1:k])
	for(j in c(1:k)){
		a[2,j] <- a[1,j]-a[2,k+1]*a[1,k-j+1]
	}
	cramerpart[k+1] <- a[2,k+1]
	a[1,] <- a[2,]
}
plot(cramerpart, type="h", xlab = "k", ylab = expression(paste("Cramer's   ",v["part"](k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)







#Example with positive dependencies

#Transition matrix Q for certain 3-state Markov chain:

Q <- rbind(c(0.9,0.05,0.25), c(0.05,0.8,0.05), c(0.05,0.15,0.7))
nostates <- dim(Q)[1] #3

#Corresponding stationary marginal distribution:
pmarg <- eigen(Q, symmetric = FALSE)$vectors[,1]
pmarg <- Re(pmarg/sum(pmarg))
pmarg
#0.6 0.2 0.2

#Gini index:
nostates/(nostates-1) * (1-sum(pmarg^2))
#0.84

#Entropy:
-1/log(nostates) * sum(pmarg*log(pmarg))
#0.8649735



#To illustrate the transition matrix, one may draw a weighted directed graph:
#setInternet2(TRUE)
#install.packages("qgraph")
#library("qgraph")

qgraph(t(Q), directed=TRUE, weighted=TRUE, labels=c(expression(s[0]),expression(s[1]),expression(s[2])), edge.color="black", edge.width=t(Q), esize=10, asize=10, fade=FALSE, layout="circle")






#MTD(1) model, i.e., Markov chain with Q, pmarg:


#Compute kappa and v for the Markov chain:
maxlag <- 15
cohenmc <- rep(0,maxlag)
vmc <- rep(0,maxlag)

sum2mc <- sum(pmarg^2)

indprobmc <- pmarg %*% t(pmarg)

Pk <- diag(1,nostates)

for(k in c(1:maxlag)){
	Pk <- Pk %*% Q #powers of transition matrix
	bivprob <- Pk %*% diag(pmarg)
	cohenmc[k] <- sum(diag(bivprob)-pmarg^2)/(1-sum2mc)
	vmc[k] <- sqrt(sum((bivprob-indprobmc)^2/indprobmc)/(nostates-1))
}
cohenmc
#[1] 0.714285714 0.512500000 0.369285714 0.267156250 0.193991964 0.141350078 0.103319027 0.075738474 0.055665839 0.041009668 0.030276439 0.022394801 0.016592958 0.012312675 0.009148678

#Plot of Cohen's kappa:
plot(cohenmc, type="h", xlab = "k", ylab = expression(paste("Cohen's   ",kappa(k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)


#Compute "partial Cohen's kappa" recursively:
cohenpart <- rep(0,maxlag)
cohenpart[1] <- cohenmc[1]

a <- array(0, c(2,maxlag))
a[1,1] <- cohenmc[1]

for(k in c(1:(maxlag-1))){
	a[2,k+1] <- (cohenmc[k+1] - a[1,1:k] %*% cohenmc[k:1]) / (1 - a[1,1:k] %*% cohenmc[1:k])
	for(j in c(1:k)){
		a[2,j] <- a[1,j]-a[2,k+1]*a[1,k-j+1]
	}
	cohenpart[k+1] <- a[2,k+1]
	a[1,] <- a[2,]
}
plot(cohenpart, type="h", xlab = "k", ylab = expression(paste("Cohen's   ",kappa["part"](k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)





vmc
# [1] 0.70769579 0.51360450 0.38048325 0.28606941 0.21712810 0.16567311 0.12670978 0.09695623 0.07414184 0.05662496 0.04318016 0.03287401 0.02498753 0.01896417 0.01437261


#Plot of Cramer's v:
plot(vmc, type="h", xlab = "k", ylab = "Cramer's   v(k)", lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)


#Compute "partial Cramer's v" recursively:
cramerpart <- rep(0,maxlag)
cramerpart[1] <- vmc[1]

a <- array(0, c(2,maxlag))
a[1,1] <- vmc[1]

for(k in c(1:(maxlag-1))){
	a[2,k+1] <- (vmc[k+1] - a[1,1:k] %*% vmc[k:1]) / (1 - a[1,1:k] %*% vmc[1:k])
	for(j in c(1:k)){
		a[2,j] <- a[1,j]-a[2,k+1]*a[1,k-j+1]
	}
	cramerpart[k+1] <- a[2,k+1]
	a[1,] <- a[2,]
}
plot(cramerpart, type="h", xlab = "k", ylab = expression(paste("Cramer's   ",v["part"](k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)



#MTD(2) model
lambda1 <- 0.2
lambda2 <- 1-lambda1


#Compute kappa and v for the Markov chain:
maxlag <- 15
cohenmtd2 <- rep(0,maxlag)
vmtd2 <- rep(0,maxlag)

sum2mc <- sum(pmarg^2)

indprobmc <- pmarg %*% t(pmarg)


#Solve YW-equations for MTD(2) using vec and Kronecker product:
P0 <- diag(pmarg) #Initialized as P(0)

#Right-hand side of equation:
rhs <- lambda1*(Q %*% P0) + lambda1*lambda2*(Q %*% t(P0) %*% t(Q))

#Dimension of vec computations:
dim.vec <- prod(dim(Q))

#Coefficient matrix for left-hand side:
lhs.coef <- diag(1, dim.vec) - lambda2^2*kronecker(Q,Q)
P1 <-  matrix( solve(lhs.coef) %*% as.vector(rhs), ncol=3) #vec of P(1)

#Cross-checking YW equation:
P1 - (lambda1 * Q %*% P0) - (lambda2 * Q %*% t(P1))


Pk1 <- P0 #Initialized as P(0)
Pk <- P1 #Initialized as P(1)
bivprob <- Pk #Pk already contains bivariate probabilities
cohenmtd2[1] <- sum(diag(bivprob)-pmarg^2)/(1-sum2mc)
vmtd2[1] <- sqrt(sum((bivprob-indprobmc)^2/indprobmc)/(nostates-1))

for(k in c(2:maxlag)){
	Pk2 <- Pk1
	Pk1 <- Pk
	Pk <- (lambda1 * Q %*% Pk1) + (lambda2 * Q %*% Pk2)
	bivprob <- Pk #Pk already contains bivariate probabilities
	cohenmtd2[k] <- sum(diag(bivprob)-pmarg^2)/(1-sum2mc)
	vmtd2[k] <- sqrt(sum((bivprob-indprobmc)^2/indprobmc)/(nostates-1))
}
cohenmtd2
#0.35077519 0.62278516 0.29496609 0.40124464 0.23036233 0.26574750 0.17329946 0.17985412 0.12774556 0.12375028 0.09310739 0.08621446 0.06744307 0.06062823 0.04869805

#Plot of Cohen's kappa:
plot(cohenmtd2, type="h", xlab = "k", ylab = expression(paste("Cohen's   ",kappa(k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)


#Compute "partial Cohen's kappa" recursively:
cohenpart <- rep(0,maxlag)
cohenpart[1] <- cohenmtd2[1]

a <- array(0, c(2,maxlag))
a[1,1] <- cohenmtd2[1]

for(k in c(1:(maxlag-1))){
	a[2,k+1] <- (cohenmtd2[k+1] - a[1,1:k] %*% cohenmtd2[k:1]) / (1 - a[1,1:k] %*% cohenmtd2[1:k])
	for(j in c(1:k)){
		a[2,j] <- a[1,j]-a[2,k+1]*a[1,k-j+1]
	}
	cohenpart[k+1] <- a[2,k+1]
	a[1,] <- a[2,]
}
plot(cohenpart, type="h", xlab = "k", ylab = expression(paste("Cohen's   ",kappa["part"](k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)




vmtd2
#0.34444854 0.61611734 0.29440472 0.40270385 0.23672447 0.27557452 0.18502870 0.19467796 0.14239722 0.14034108 0.10849319 0.10239056 0.08204969 0.07519631 0.06168374

#Plot of Cramer's v:
plot(vmtd2, type="h", xlab = "k", ylab = "Cramer's   v(k)", lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)


#Compute "partial Cramer's v" recursively:
cramerpart <- rep(0,maxlag)
cramerpart[1] <- vmtd2[1]

a <- array(0, c(2,maxlag))
a[1,1] <- vmtd2[1]

for(k in c(1:(maxlag-1))){
	a[2,k+1] <- (vmtd2[k+1] - a[1,1:k] %*% vmtd2[k:1]) / (1 - a[1,1:k] %*% vmtd2[1:k])
	for(j in c(1:k)){
		a[2,j] <- a[1,j]-a[2,k+1]*a[1,k-j+1]
	}
	cramerpart[k+1] <- a[2,k+1]
	a[1,] <- a[2,]
}
plot(cramerpart, type="h", xlab = "k", ylab = expression(paste("Cramer's   ",v["part"](k))), lwd=4, xlim=c(0,15), ylim=c(-1,1), xaxp=c(0,15,3))
abline(h=0)


