#Fitting DAR(p) and MTD(p) models to a categorical time series



#Conditional log-likelihood function for DAR(p) process (par=c(parmarg,pardep) with parmarg=c(pi1,...,pid)):

ll_darp <- function(par,databin){
	T <- dim(databin)[1]
	d <- dim(databin)[2]-1
	p <- length(par)-d
	parmarg <- par[1:d]
	pardep <- par[(d+1):(d+p)]
	phi0 <- 1-sum(pardep)
	probpi <- c(1-sum(parmarg), parmarg)
	
	value <- 0
	for (t in c((p+1):T)) {
		krondeltas <- rep(0,p)
		for(r in c(1:p)){
			krondeltas[r] <- (databin[t,] %*% databin[t-r,])[1,1]
		}
		
		value<-value-log(phi0* (probpi %*% databin[t,])[1,1] + (pardep %*% krondeltas)[1,1])
		}
	value
}
#ll_darp(c(0.331,0.208,0.241, 0.081),databin)
#11444.26
#ll_darp(c(0.331,0.209,0.243, 0.079,0.020),databin)
#11437.76


#Constraints for optimization A %*% par >= b:
#all parameters nonnegative, sums <= 1
Amat <- function(nostates,p){
	value <- array(0,c((nostates+p+1), (nostates+p-1)))
	value[1:(nostates+p-1),1:(nostates+p-1)] <- diag(1, (nostates+p-1))
	value[(nostates+p),c(1:(nostates-1))] <- rep(-1, (nostates-1))
	value[(nostates+p+1),c(nostates:(nostates+p-1))] <- rep(-1, p)
	value
}

bvec <- function(nostates,p){
	value <- rep(0, (nostates+p+1))
	value[(nostates+p):(nostates+p+1)] <- rep(-1, 2)
	value
}



#Conditional log-likelihood function for Markov chain with transition matrix Q (par=c(Qcol1,Qcol2,...) without probabilities p(0|j) of length l=d*(d+1), i.e., 4*l+1=(2*d+1)^2):

ll_mc <- function(par,datanum){
	T <- length(datanum)
	nostates <- (sqrt(1+4*length(par))+1)/2
	
	Qmat <- array(0, c(nostates,nostates))
	for(r in c(1:nostates)){
		qcol <- par[((r-1)*(nostates-1)+1):(r*(nostates-1))]
		Qmat[,r] <- c(1-sum(qcol),qcol) #add p(0|j)
	}

	
	value <- 0
	for (t in c(2:T)) {
		value<-value-log(Qmat[datanum[t],datanum[t-1]])
		}
	value
}
#ll_mc(rep(hatpi[2:4],4),datanum)
#11522.11


#Constraints for optimization A %*% par >= b:
#all parameters nonnegative, sums <= 1

AMCmat <- function(nostates){
	value <- array(0,c((nostates^2), (nostates*(nostates-1))))
	value[1:(nostates*(nostates-1)),1:(nostates*(nostates-1))] <- diag(1, (nostates*(nostates-1)))
	for(r in c(1:nostates)){
		value[(nostates*(nostates-1)+r), ((r-1)*(nostates-1)+1):(r*(nostates-1))] <- rep(-1, (nostates-1))
	}
	value
}

bMCvec <- function(nostates){
	value <- rep(0, (nostates^2))
	value[(nostates*(nostates-1)+1):(nostates^2)] <- rep(-1, nostates)
	value
}



#Conditional log-likelihood function for MTD(p) with p>=2 with transition matrix Q (par=c(Qcol1,Qcol2,...,lambda2,...,lambdap) without probabilities p(0|j), par is of length l=d*(d+1) + p-1, i.e., 4*(l-p+1)+1=(2*d+1)^2):

ll_mtdp <- function(par,datanum,orderp){
	T <- length(datanum)
	p <- orderp
	nostates <- (sqrt(1+4*(length(par)-p+1))+1)/2
	
	Qmat <- array(0, c(nostates,nostates))
	for(r in c(1:nostates)){
		qcol <- par[((r-1)*(nostates-1)+1):(r*(nostates-1))]
		Qmat[,r] <- c(1-sum(qcol),qcol) #add p(0|j)
	}
	lambda <- par[(nostates*(nostates-1)+1):(nostates*(nostates-1)+p-1)] #lambda2,...,lambdap
	lambda <- c(1-sum(lambda), lambda) #add lambda1
	
	value <- 0
	for (t in c((p+1):T)) {
		cp <- 0
		for(k in c(1:p)){
			cp <- cp + lambda[k]*Qmat[datanum[t],datanum[t-k]]
		}
		value <- value-log(cp)
		}
	value
}
#ll_mtdp(c(mc$par,0),datanum,2)
#11355.28
#ll_mc(mc$par,datanum[2:Tlen])
#11355.28

#Constraints for optimization A %*% par >= b:
#all parameters nonnegative, sums <= 1

AMTDmat <- function(nostates,p){
	value <- cbind(AMCmat(nostates), array(0,c((nostates^2), (p-1))))
	value <- rbind(value, array(0,c(p, (nostates*(nostates-1)+p-1))))
	
	for(r in c(1:(p-1))){
		value[nostates^2+r,nostates*(nostates-1)+r] <- 1
	}
	value[(nostates^2+p),c((nostates*(nostates-1)+1):(nostates*(nostates-1)+p-1))] <- rep(-1, p-1)
	value
}

bMTDvec <- function(nostates,p){
	value <- c(bMCvec(nostates), rep(0,p))
	value[(nostates^2+p)] <- -1
	value
}







#Genome of the Bovine leukemia virus, see Weiß & Göb (2008).
#
# four possible states (nominal; shown in lexicographic order):
# a: adenine;
# c: cytosine;
# g: guanine;
# t: thymine.

data <- scan(file.choose(), what="character") #Bovine.txt"
Tlen <- length(data)
Tlen #8419

states <- c("a","c","g","t") #position in states corresponds to numeric code
nostates <- length(states)
nostates #4


#Numeric coding by 1-4:
datanum <- match(data, states)

#Binarization:
bincodes <- diag(1,nostates)
databin <- bincodes[datanum,] #assign row vectors

#relative frequencies
hatpi <- colMeans(databin)
hatpi
# 0.2197411 0.3313933 0.2102387 0.2386269


#Rate evolution graph
cols <- gray((0:(nostates-1))/nostates)

matplot(apply(databin, 2, cumsum), type="b", pch=19, cex=0.1, xlab = "t", ylab = "Genome Bovine leukemia", col=cols)
legend("topleft", states, col=cols, text.col=cols, title = expression("DNA bases:"), pch=19, cex=0.8, y.intersp=1.2)





maxlag <- 15
hatbivprob <- array(0,c(nostates,nostates,maxlag))

for(k in c(1:maxlag)){
	for(i in c(1:nostates)){
		for(j in c(1:nostates)){
			hatbivprob[i,j,k] <- mean(databin[(k+1):Tlen,i]*databin[1:(Tlen-k),j])
		}
	}
}

#... which is to be compared with
indprob <- hatpi %*% t(hatpi)


#Cramer's v
cramer <- rep(0,maxlag)
for(k in c(1:maxlag)){
	cramer[k] <- sqrt(sum((hatbivprob[,,k]-indprob)^2/indprob)/(nostates-1))
}
#cramer
#0.11344728 0.04447183 0.02808013 0.02221853 0.02941132 0.03287870 0.02028634 0.02071309 0.02085900 0.01327671 0.02301690 0.02165745 0.01144901 0.02268569 0.02837755

#Critical value
alpha <- 0.05
crit <- sqrt(qchisq(1-alpha,(nostates-1)^2)/Tlen/(nostates-1))
#0.0258819

#Plot of Cramer's v:
plot(cramer, type="h", xlab = "k", ylab = "Cramer's   v(k)", lwd=4, ylim=c(0,0.12))
abline(h=crit, col=gray(0.5), lwd=2, lty=2)
abline(h=0)


#sums of powers of probabilities:
sum2 <- sum(hatpi^2)
sum3 <- sum(hatpi^3)

#lower limit of range of kappa
-sum2/(1-sum2)
#-0.3499845

#Cohen's kappa
cohen <- rep(0,maxlag)
for(k in c(1:maxlag)){
	cohen[k] <- sum(diag(hatbivprob[,,k])-hatpi^2)/(1-sum2)
}
#cohen
#0.0804453171  0.0248418712  0.0008253725 -0.0065125335 -0.0151357440  0.0200456969 -0.0002916790  0.0076145076  0.0127937114  0.0035255077  0.0059759452  0.0093904373  0.0071848213  0.0054605309  0.0032539223



#Critical value
crit2 <- qnorm(1-alpha/2, sd=sqrt((1-(1+2*sum3-3*sum2)/(1-sum2)^2)/Tlen))
#0.01245973

#Plot of Cohen's kappa:
plot(cohen, type="h", xlab = "k", ylab = expression(paste("Cohen's   ",kappa(k))), lwd=4, ylim=c(-0.02,0.10))
abline(h=c(-1/Tlen+crit2,-1/Tlen-crit2), col=gray(0.5), lwd=2, lty=2)
abline(h=-1/Tlen)






#An inspection of Cohen's kappa or Cramer's v suggests to fit a DAR(1) or DAR(2) model to the data. To derive some kind of "moment estimates", we compute:
cohen[1:2]
#0.08044532 0.02484187



#First fit DAR(1) model:

#Initialize DAR(1) estimates:
start <- c(hatpi[2:nostates], cohen[1])

dar1 <- suppressWarnings(constrOptim(start, ll_darp, NULL, ui=Amat(nostates,1), ci=bvec(nostates,1), databin=databin))
dar1$par
#0.33090108 0.20837214 0.24106708 0.08058538
dar1$value
#11444.26



#Marginal distribution of fitted DAR(1) model:
pmargMLdar1 <- dar1$par[1:(nostates-1)]
pmargMLdar1 <- c(1-sum(pmargMLdar1), pmargMLdar1)
pmargMLdar1
#0.2196597 0.3309011 0.2083721 0.2410671

#Compare dispersion behaviour:

#Gini index:
nostates/(nostates-1) * (1-sum(hatpi^2))
#0.9876656
nostates/(nostates-1) * (1-sum(pmargMLdar1^2))
#0.9876291

#Entropy:
-1/log(nostates) * sum(hatpi*log(hatpi))
#0.9873583
-1/log(nostates) * sum(pmargMLdar1*log(pmargMLdar1))
#0.9872843


#Transition matrix of fitted DAR(1) model:
phi1MLdar1 <- dar1$par[nostates]
TMdar1 <- array(pmargMLdar1, c(nostates,nostates)) *(1-phi1MLdar1) + diag(phi1MLdar1, nostates)
TMdar1
          # [,1]      [,2]      [,3]      [,4]
# [1,] 0.2825437 0.2019583 0.2019583 0.2019583
# [2,] 0.3042353 0.3848207 0.3042353 0.3042353
# [3,] 0.1915804 0.1915804 0.2721658 0.1915804
# [4,] 0.2216406 0.2216406 0.2216406 0.3022260



#h-step-ahead forecast distribution based on fitted DAR(1) model
cprop <- function(j,h){ #h-step-ahead given j
	value <- rep(0, nostates)
	for(i in c(1:nostates)){
		value[i] <- pmargMLdar1[i]*(1-phi1MLdar1^h)
		if(i==j) value[i] <- value[i] + phi1MLdar1^h
	}
	value
}
#1-step-ahead, conditional on `a' (=1):
cprop(1,1) #0.2825437 0.3042353 0.1915804 0.2216406
#1-step-ahead, conditional on `g' (=3):
cprop(3,1) #0.2019583 0.3042353 0.2721658 0.2216406

#2-step-ahead, conditional on `a' (=1):
cprop(1,2) #0.2247272 0.3287522 0.2070190 0.2395016
#2-step-ahead, conditional on `g' (=3):
cprop(3,2) #0.2182332 0.3287522 0.2135130 0.2395016

#3-step-ahead, conditional on `a' (=1):
cprop(1,3) #0.2200681 0.3307279 0.2082631 0.2409409
#3-step-ahead, conditional on `g' (=3):
cprop(3,3) #0.2195447 0.3307279 0.2087864 0.2409409

#Always `c' as point forecast.





#Next, fit DAR(2) model:

#Initialize DAR(2) estimates:
#"moment estimates"
phi1MM <- cohen[1]*(1-cohen[2])/(1-cohen[1]^2) #0.07895788
phi2MM <- (cohen[2]-cohen[1]^2)/(1-cohen[1]^2) #0.01849008
#Also the following could be used here:
#acf2AR(c(1,cohen[1:2]))[2,]

start <- c(hatpi[2:nostates], phi1MM,phi2MM)

dar2 <- suppressWarnings(constrOptim(start, ll_darp, NULL, ui=Amat(nostates,2), ci=bvec(nostates,2), databin=databin))
dar2$par
#0.3311943 0.2089206 0.2409296 0.0791415 0.0200846
dar2$value
#11437.63





#Fit full 1st order Markov chain (=MTD(1) model)

mc <- suppressWarnings(constrOptim(rep(hatpi[2:nostates], nostates), ll_mc, NULL, ui=AMCmat(nostates), ci=bMCvec(nostates), datanum=datanum))
mc$par
#0.2681649 0.2261211 0.2237690 0.3917552 0.1335573 0.2722630 0.2974168 0.2977907 0.1718268 0.3356005 0.2248259 0.2700155
mc$value
#11356.73


#Compute transition matrix
TMmc <- array(0, c(nostates,nostates))
for(r in c(1:nostates)){
	qcol <- mc$par[((r-1)*(nostates-1)+1):(r*(nostates-1))]
	TMmc[,r] <- c(1-sum(qcol),qcol) #add p(0|j)
}
TMmc
          # [,1]      [,2]      [,3]      [,4]
# [1,] 0.2819450 0.2024246 0.2329657 0.1695581
# [2,] 0.2681649 0.3917552 0.2974168 0.3356005
# [3,] 0.2261211 0.1335573 0.2977907 0.2248259
# [4,] 0.2237690 0.2722630 0.1718268 0.2700155
#
#Can be used for 1-step-ahead forecasting

#Corresponding stationary marginal distribution:
pmarg <- eigen(TMmc, symmetric = FALSE)$vectors[,1]
pmarg <- Re(pmarg/sum(pmarg))
pmarg
#0.2183159 0.3314656 0.2101929 0.2400255

#2-step-ahead forecasting:
TMmc %*% TMmc
          # [,1]      [,2]      [,3]      [,4]
# [1,] 0.2243965 0.2136521 0.2243976 0.2138999
# [2,] 0.3230120 0.3388491 0.3252209 0.3344271
# [3,] 0.2172149 0.1990781 0.2197110 0.2108201
# [4,] 0.2353767 0.2484207 0.2306704 0.2408529
#Columns now already close to marginal distribution,
#mode forecast always `c'.

#This is not surprising, since the second largest eigenvalue is very small:

eigen(TMmc, symmetric=FALSE)$values
#[1] 1.00000000+0.0000000i 0.15621412+0.0000000i 0.04264612+0.0459932i 0.04264612-0.0459932i


#Compute kappa and v for the Markov chain:
maxlag <- 15
cohenmc <- rep(0,maxlag)
vmc <- rep(0,maxlag)

sum2mc <- sum(pmarg^2)

indprobmc <- pmarg %*% t(pmarg)

Pk <- diag(1,nostates)

for(k in c(1:maxlag)){
	Pk <- Pk %*% TMmc #powers of transition matrix
	bivprob <- Pk %*% diag(pmarg)
	cohenmc[k] <- sum(diag(bivprob)-pmarg^2)/(1-sum2mc)
	vmc[k] <- sqrt(sum((bivprob-indprobmc)^2/indprobmc)/(nostates-1))
}
cohenmc
#8.031310e-02 8.065702e-03 1.176438e-03 1.942770e-04 3.157309e-05 4.995273e-06 7.808987e-07 1.217874e-07 1.900563e-08 2.968090e-09 4.636603e-10 7.243393e-11 1.131554e-11 1.767701e-12 2.761785e-13
vmc
#1.130852e-01 1.623882e-02 2.462517e-03 3.802798e-04 5.915033e-05 9.231440e-06 1.442283e-06 2.253545e-07 3.520692e-08 5.499907e-09 8.591576e-10 1.342117e-10 2.096573e-11 3.275143e-12 5.116337e-13

#Compare dispersion behaviour:

#Gini index:
nostates/(nostates-1) * (1-sum(hatpi^2))
#0.9876656
nostates/(nostates-1) * (1-sum(pmarg^2))
#0.9875672
nostates/(nostates-1) * (1-sum(pmargMLdar1^2))
#0.9876291

#Entropy:
-1/log(nostates) * sum(hatpi*log(hatpi))
#0.9873583
-1/log(nostates) * sum(pmarg*log(pmarg))
#0.9872459
-1/log(nostates) * sum(pmargMLdar1*log(pmargMLdar1))
#0.9872843






#Fit MTD(2) model

mtd2 <- suppressWarnings(constrOptim(c(mc$par,0.05), ll_mtdp, NULL, ui=AMTDmat(nostates,2), ci=bMTDvec(nostates,2), datanum=datanum, orderp=2))
mtd2$par
#2.676117e-01 2.259987e-01 2.235553e-01 3.915645e-01 1.347078e-01 2.718472e-01 2.953544e-01 2.981864e-01 1.729255e-01 3.382476e-01 2.253260e-01 2.670363e-01 4.787173e-06
mtd2$value
#11355.21


#Matrix Q of MTD(2) model:
TMmtd2 <- array(0, c(nostates,nostates))
for(r in c(1:nostates)){
	qcol <- mtd2$par[((r-1)*(nostates-1)+1):(r*(nostates-1))]
	TMmtd2[,r] <- c(1-sum(qcol),qcol) #add p(0|j)
}
TMmtd2
          # [,1]      [,2]      [,3]      [,4]
# [1,] 0.2828344 0.2018806 0.2335337 0.1693902
# [2,] 0.2676117 0.3915645 0.2953544 0.3382476
# [3,] 0.2259987 0.1347078 0.2981864 0.2253260
# [4,] 0.2235553 0.2718472 0.1729255 0.2670363

#Comparison with matrix Q of MTD(1) model:
TMmc
          # [,1]      [,2]      [,3]      [,4]
# [1,] 0.2819450 0.2024246 0.2329657 0.1695581
# [2,] 0.2681649 0.3917552 0.2974168 0.3356005
# [3,] 0.2261211 0.1335573 0.2977907 0.2248259
# [4,] 0.2237690 0.2722630 0.1718268 0.2700155



lambdamtd2 <- mtd2$par[(nostates*(nostates-1)+1)] #lambda2
lambdamtd2 <- c(1-sum(lambdamtd2), lambdamtd2) #add lambda1
lambdamtd2
#9.999952e-01 4.787173e-06

#No visible improvement compared to MTD(1) model.






#Finally, compare BICs (and correct for conditioning):
#DAR(1):
2*Tlen/(Tlen-1)*dar1$value+log(Tlen)*(nostates)
#22927.38

#DAR(2):
2*Tlen/(Tlen-2)*dar2$value+log(Tlen)*(nostates+1)
#22925.89

#MC:
2*Tlen/(Tlen-1)*mc$value+log(Tlen)*(nostates*(nostates-1))
#22824.62

#MTD(2):
2*Tlen/(Tlen-2)*mtd2$value+log(Tlen)*(nostates*(nostates-1)+1)
#22833.32





#Exercise: CIs for Gini and entropy based on CML-fitted DAR(1) model
const <- (1+phi1MLdar1)/(1-phi1MLdar1)
const
#1.175297

#Gini:
gini <- nostates/(nostates-1) * (1-sum(hatpi^2))
gini
#0.9876656

#Gini:
ginidar1 <- nostates/(nostates-1) * (1-sum(pmargMLdar1^2))
ginidar1
#0.9876291

#Exact bias correction:
bias1 <- (1/( 1-1/Tlen*(1+2*( (1-c(1:(Tlen-1))/Tlen) %*% (phi1MLdar1^c(1:(Tlen-1))) )) ))[1,1]
bias1
#1.00014
bias1*gini
#0.9878035

#Approximate bias correction:
bias2 <- Tlen/(Tlen-const)
bias2
#1.00014
bias2*gini
#0.9878035

#Standard error:
SEgini <- sqrt( 4*const/Tlen*nostates/(nostates-1) * (sum3-sum2^2) )
SEgini
#0.001411068

#Confidence interval:
z <- qnorm(0.975)
bias1*gini-z*SEgini #0.9850379
bias1*gini+z*SEgini #0.9905692


#Entropy:
entropy <- -1/log(nostates) * sum(hatpi*log(hatpi))
entropy
#0.9873583

entropydar1 <- -1/log(nostates) * sum(pmargMLdar1*log(pmargMLdar1))
entropydar1
#0.9872843

#Standard error:
SEentropy <- sqrt( const/Tlen/(log(nostates)^2) * (sum(hatpi*(log(hatpi)^2))-(sum(hatpi*log(hatpi)))^2) )
SEentropy
#0.001628307

#Confidence interval:
entropy-z*SEentropy #0.9841669
entropy+z*SEentropy #0.9905497








#Exercise: approximate CI for phi1 according to Corollary 5.3.1 in Weiß (2013):
phi1 <- cohen[1]
phi1
#0.08044532

#Approximate standard error:
SEphi1 <- sqrt( 1/Tlen*(2*(2*phi1-1)*(1-phi1)* (sum3-sum2^2)/(1-sum2)^2 + (1-phi1)*sum2/(1-sum2) + phi1*(1-phi1)) )
SEphi1
#0.006791155

#Confidence interval:
phi1-z*SEphi1 #0.0671349
phi1+z*SEphi1 #0.09375574




